#include "Ase.h"



/*****************************************************************************
* 
*****************************************************************************/
#define ASE_PACKET_TYPE(pid, cnt, dst) (pid | cnt << 2 | dst)
#define ASE_STOP_BYTE(cnt, dst) (0x60 | cnt << 2 | dst)

#define BULK_BUFFER_SIZE        300 

#define T0_DATA_CHUNK_SIZE           28
#define T1_DATA_CHUNK_SIZE           32


/*****************************************************************************
* 
*****************************************************************************/
static void GetDefaultReaderParams (reader* globalData, struct card_params* params) {
    params->protocol = 0;
    params->N = 0x03;
    params->CWT[0] = 0x00;
    params->CWT[1] = 0x85;
    params->CWT[2] = 0xF0;
    params->BWT[0] = 0x00;
    params->BWT[1] = 0x38;
    params->BWT[2] = 0xA4;
    if (globalData->dataMemory[3] == 9 && globalData->dataMemory[4] == 0) {
        params->A = 0x0C;
        params->B = 0x00;
    }
    else {
        params->A = 0x01;
        params->B = 0x74;
    }
    params->freq = 0x02;
}




/*****************************************************************************
*
*****************************************************************************/
/* returns 0 on success */
int cardCommandInit (reader* globalData, char socket, char needToBePoweredOn) {
    if (!globalData->cards[socket].status) {
        return ASE_READER_NO_CARD_ERROR;
    }

    if (needToBePoweredOn && globalData->cards[socket].status != 2) {
        return ASE_READER_CARD_NOT_POWERED_ERROR;
    }

    globalData->curPos = globalData->lastPos = 0;

    return ASE_OK;
}


/*****************************************************************************
*
*****************************************************************************/
int readerCommandInit (reader* globalData, char needToBeStarted) {
    if (needToBeStarted && !globalData->readerStarted) {
        return ASE_ERROR_COMMAND;
    }

    globalData->curPos = globalData->lastPos = 0;

    return ASE_OK;
}


/*****************************************************************************
*
*****************************************************************************/
void lock_mutex (reader* globalData) {
    int retVal;

#ifdef ASE_DEBUG
    /*    syslog(LOG_INFO, "  Trying...");*/
#endif
    
    retVal = pthread_mutex_lock(&(globalData->semaphore)); 
    if (retVal != 0) {
#ifdef ASE_DEBUG
        syslog(LOG_INFO, "Error! - in pthread_mutex_lock\n");
#endif
    }
    
#ifdef ASE_DEBUG
    /*    syslog(LOG_INFO, "Entering mutex\n");*/
#endif
}


/*****************************************************************************
*
*****************************************************************************/
void unlock_mutex (reader* globalData) {
    int retVal;

#ifdef ASE_DEBUG
    /*    syslog(LOG_INFO, "  Trying...");*/
#endif
    
    retVal = pthread_mutex_unlock(&(globalData->semaphore)); 
    if (retVal != 0) {
#ifdef ASE_DEBUG
        syslog(LOG_INFO, "Error! - in pthread_mutex_unlock\n");
#endif
    }
    
#ifdef ASE_DEBUG
    /*    syslog(LOG_INFO, "Leaving mutex\n");*/
#endif
}


/*****************************************************************************
*
*****************************************************************************/
int ReaderStartup (reader* globalData, char* response, int* len) {
    char cmd[5];
    int retVal, retVal2, i;

#ifdef ASE_DEBUG
    syslog(LOG_INFO, "\n ReaderStartup - Enter\n");
#endif

    if (globalData->readerStarted) {
        return ASE_ERROR_COMMAND;
    }

    globalData->readerStarted = 0;
    globalData->commandCounter = 1;
    globalData->cards[0].status = globalData->cards[0].ceei = globalData->cards[0].atr.length = 0;
    globalData->cards[1].status = globalData->cards[1].ceei = globalData->cards[1].atr.length = 0;
    globalData->cards[0].cwt = globalData->cards[1].cwt = 1500000; /* 1.5 sec */
    globalData->curPos = globalData->lastPos = globalData->curEvent = globalData->lastEvent = 0;

    /* init mutex */
    retVal = pthread_mutex_init(&(globalData->semaphore), NULL); 
    if (retVal != 0)
        return ASE_ERROR_COMMAND; //??????????

    if (retVal = readerCommandInit(globalData, 0))
        return retVal;

    cmd[0] = ASE_PACKET_TYPE(0x50, globalData->commandCounter, 0);
    globalData->commandCounter++;
    globalData->commandCounter %= 4;
    cmd[1] = 0x10;
    cmd[2] = 0x01;
    cmd[3] = 0x00;
    cmd[4] = cmd[0] ^ cmd[1] ^ cmd[2] ^ cmd[3];

    lock_mutex(globalData);
    retVal = sendCloseResponseCommand(globalData, cmd, 5, response, len);
    unlock_mutex(globalData);

    if (retVal < 0) {
        return retVal; 
    }

    globalData->readerStarted = 1;

    /* copy response to the ASE data memory field */
    for (i = 0 ; i < ASE_DATA_MEMORY_SIZE ; ++i)
        globalData->dataMemory[i] = response[i];

#ifdef ASE_DEBUG
    syslog(LOG_INFO, " ReaderStartup - Exit\n");
#endif

    return ASE_OK; 
}



/*****************************************************************************
*
*****************************************************************************/
int ReaderFinish (reader* globalData) {
    char cmd[4], ack;
    int retVal, retVal2, actual;

#ifdef ASE_DEBUG
    syslog(LOG_INFO, "\n ReaderFinish - Enter\n");
#endif

    if (retVal = readerCommandInit(globalData, 1))
        return retVal;

    cmd[0] = ASE_PACKET_TYPE(0x50, globalData->commandCounter, 0);
    globalData->commandCounter++;
    globalData->commandCounter %= 4;
    cmd[1] = 0x11;
    cmd[2] = 0x0;
    cmd[3] = cmd[0] ^ cmd[1] ^ cmd[2];

    lock_mutex(globalData);
    retVal = sendControlCommand(globalData, cmd, 4, &ack, &actual);
    unlock_mutex(globalData);

    if (retVal < 0) {
        return retVal; 
    }

    if (ack != 0x20) {
        return ackStatus(ack);
    }

    globalData->readerStarted = 0;

    /* destroy mutex */
    retVal = pthread_mutex_destroy(&(globalData->semaphore)); 
    if (retVal != 0)
        return ASE_ERROR_COMMAND; //??????????

#ifdef ASE_DEBUG
    syslog(LOG_INFO, " ReaderFinish - Exit\n");
#endif

    return ASE_OK; 
}


/*****************************************************************************
*
*****************************************************************************/
int GetFIDITable (reader* globalData, char* response, int* len) {
    char cmd[4];
    int retVal, retVal2, i;

#ifdef ASE_DEBUG
    syslog(LOG_INFO, "\n GetFIDITable - Enter\n");
#endif

    if (retVal = readerCommandInit(globalData, 1))
        return retVal;

    cmd[0] = ASE_PACKET_TYPE(0x50, globalData->commandCounter, 0);
    globalData->commandCounter++;
    globalData->commandCounter %= 4;
    cmd[1] = 0x12;
    cmd[2] = 0x00;
    cmd[3] = cmd[0] ^ cmd[1] ^ cmd[2];

    lock_mutex(globalData);
    retVal = sendCloseResponseCommand(globalData, cmd, 4, response, len);
    unlock_mutex(globalData);

    if (retVal < 0) {
        return retVal; 
    }

    /* save the table */
    globalData->fidiTable.len = *len;
    for (i = 0 ; i < *len ; ++i) 
        globalData->fidiTable.table[i] = response[i]; /* Fi, Di */

#ifdef ASE_DEBUG
    syslog(LOG_INFO, " GetFIDITable - Exit\n");
#endif

    return ASE_OK; 
}


/*****************************************************************************
*
*****************************************************************************/
int GetStatus (reader* globalData, char* response, int* len) {
    char cmd[4];
    int retVal, retVal2;

#ifdef ASE_DEBUG
    syslog(LOG_INFO, "\n GetStatus - Enter\n"); 
#endif

    if (retVal = readerCommandInit(globalData, 1))
        return retVal;

    cmd[0] = ASE_PACKET_TYPE(0x50, globalData->commandCounter, 0);
    globalData->commandCounter++;
    globalData->commandCounter %= 4;
    cmd[1] = 0x16;
    cmd[2] = 0x0;
    cmd[3] = cmd[0] ^ cmd[1] ^ cmd[2];

    lock_mutex(globalData);
    retVal = sendCloseResponseCommand(globalData, cmd, 4, response, len);
    unlock_mutex(globalData);

    if (retVal < 0) {
        return retVal; 
    }

    if (response[0] & 0x01) {
        if (!(globalData->cards[0].status)) /* only if not powered on */
	       globalData->cards[0].status = 1; 
    }
    else
	    globalData->cards[0].status = 0;

    if (response[0] & 0x02) {
        if (!(globalData->cards[1].status)) /* only if not powered on */
	        globalData->cards[1].status = 1;
    }
    else
	    globalData->cards[1].status = 0;

#ifdef ASE_DEBUG
    syslog(LOG_INFO, " GetStatus - Exit\n"); 
#endif

    return ASE_OK; 
}


/*****************************************************************************
*
*****************************************************************************/
int SetCardParameters (reader* globalData, char socket, struct card_params params) {
    char cmd[15], ack;
    int retVal, i, actual, retVal2;

#ifdef ASE_DEBUG
    syslog(LOG_INFO, "\n SetCardParameters - Enter\n");
#endif

    if (retVal = cardCommandInit(globalData, socket, 0))
        return retVal;

    cmd[0] = ASE_PACKET_TYPE(0x50, globalData->commandCounter, socket);
    globalData->commandCounter++;
    globalData->commandCounter %= 4;
    cmd[1] = 0x15;
    cmd[2] = 0x0B;
    cmd[3] = params.protocol;
    cmd[4] = params.N;
    cmd[5] = params.CWT[0];
    cmd[6] = params.CWT[1];
    cmd[7] = params.CWT[2];
    cmd[8] = params.BWT[0];
    cmd[9] = params.BWT[1];
    cmd[10] = params.BWT[2];
    cmd[11] = params.A;
    cmd[12] = params.B;
    cmd[13] = params.freq;
    cmd[14] = cmd[0];
    for (i = 1 ; i < 14 ; i++)
        cmd[14] ^= cmd[i];


    lock_mutex(globalData);
    retVal = sendControlCommand(globalData, cmd, 15, &ack, &actual);
    unlock_mutex(globalData);

    if (retVal < 0) {
        return retVal; 
    }

    if (ack != 0x20) {
        return ackStatus(ack); 
    }

    globalData->cards[socket].cardParams = params;

#ifdef ASE_DEBUG
    syslog(LOG_INFO, " SetCardParameters - Exit\n");
#endif

    return ASE_OK; 
}



/*****************************************************************************
*
*****************************************************************************/
static int CardPowerOnOpenResponseProccessor   (reader* globalData, char socket, unsigned long arg, uchar* cs,
                                                char* outBuf, int* outBufLen) {
    int retVal, i;
    ATR* scn = &(globalData->cards[socket].atr);

    retVal = ReadATR(globalData, socket);

    if (retVal < 0) {
        return retVal; 
    }

#ifdef ASE_DEBUG
    syslog(LOG_INFO, "CardPowerOnOpenResponseProccessor - read %d bytes of ATR\n", scn->length);
#endif
    for (i = 0 ; i < scn->length ; i++) {
#ifdef ASE_DEBUG
    	syslog(LOG_INFO, "0x%x%x ", PRINT_CHAR(scn->data[i]));
#endif
    	(*cs) ^= scn->data[i];
    }
#ifdef ASE_DEBUG
    syslog(LOG_INFO, "\n");
#endif

    return ASE_OK;
}


/*****************************************************************************
*
*****************************************************************************/
int CardPowerOn (reader* globalData, char socket, uchar cardType, uchar voltage) {
    char cmd[6], response[300];
    int retVal, retVal2, len;
    struct card_params params;
    ATR* atr = &(globalData->cards[socket].atr);

#ifdef ASE_DEBUG
    syslog(LOG_INFO, "\n CardPowerOn - Enter\n");
#endif

    if (retVal = cardCommandInit(globalData, socket, 0))
        return retVal;

    GetDefaultReaderParams(globalData, &params);
    retVal = SetCardParameters(globalData, socket, params);
    if (retVal < 0) {
        return retVal; 
    }

    cmd[0] = ASE_PACKET_TYPE(0x50, globalData->commandCounter, socket);
    globalData->commandCounter++;
    globalData->commandCounter %= 4;
    cmd[1] = 0x20;
    cmd[2] = 0x02;
    cmd[3] = cardType;
    cmd[4] = voltage; 
    cmd[5] = cmd[0] ^ cmd[1] ^ cmd[2] ^ cmd[3] ^ cmd[4];

    /* CPU cards -> an open response should be receievd */
    if (cardType == 0x00 || cardType == 0x01) {
	    lock_mutex(globalData);
        retVal = sendOpenResponseCommand(globalData, socket, cmd, 6, response, &len, 
                                         CardPowerOnOpenResponseProccessor, 0, 1);
	    unlock_mutex(globalData);
    }
    /* 2/3BUS memory cards -> a close response should be received */
    else if (cardType == 0x11 || cardType == 0x12) {
        memset(atr, 0x00, sizeof(ATR));

	    lock_mutex(globalData);
        retVal = sendCloseResponseCommand(globalData, cmd, 6, response, &len); 
	    unlock_mutex(globalData);

        if (retVal == ASE_OK) {
            if (len) {
                memcpy(atr->data, response, len);
                atr->length = len;
            }
        }
    }
    /* X/I2C memory cards -> an acknowledgement should be received */
    else {
        memset(atr, 0x00, sizeof(ATR));

	    lock_mutex(globalData);
        retVal = sendControlCommand(globalData, cmd, 6, response, &len); 
	    unlock_mutex(globalData);
    }

    if (retVal < 0) {
        return retVal; 
    }

#ifdef ASE_DEBUG
    syslog(LOG_INFO, " CardPowerOn - Exit\n");
#endif

    return ASE_OK; 
}


/*****************************************************************************
*
*****************************************************************************/
int CPUCardReset (reader* globalData, char socket) {
    char cmd[4], response[300];
    int retVal, retVal2, i, len;
    struct card_params params;

#ifdef ASE_DEBUG
    syslog(LOG_INFO, "\n CPUCardReset - Enter\n");
#endif

    if (retVal = cardCommandInit(globalData, socket, 1))
        return retVal;

    GetDefaultReaderParams(globalData, &params);
    retVal = SetCardParameters(globalData, socket, params);
    if (retVal < 0) {
        return retVal; 
    }

    cmd[0] = ASE_PACKET_TYPE(0x50, globalData->commandCounter, socket);
    globalData->commandCounter++;
    globalData->commandCounter %= 4;
    cmd[1] = 0x22;
    cmd[2] = 0x00;
    cmd[3] = cmd[0] ^ cmd[1] ^ cmd[2];

    lock_mutex(globalData);
    retVal = sendOpenResponseCommand(globalData, socket, cmd, 4, response, &len, 
                                     CardPowerOnOpenResponseProccessor, 0, 1);
    unlock_mutex(globalData);

    if (retVal < 0) {
        return retVal; 
    }

#ifdef ASE_DEBUG
    syslog(LOG_INFO, " CPUCardReset - Exit\n");
#endif

    return ASE_OK; 
}


/*****************************************************************************
*
*****************************************************************************/
int CardPowerOff (reader* globalData, char socket) {
    char cmd[4], ack;
    int retVal, retVal2, actual;

#ifdef ASE_DEBUG
    syslog(LOG_INFO, "\n CardPowerOff - Enter\n");
#endif

    if (retVal = cardCommandInit(globalData, socket, 1))
        return retVal;

    cmd[0] = ASE_PACKET_TYPE(0x50, globalData->commandCounter, socket);
    globalData->commandCounter++;
    globalData->commandCounter %= 4;
    cmd[1] = 0x21;
    cmd[2] = 0x0;
    cmd[3] = cmd[0] ^ cmd[1] ^ cmd[2];

    lock_mutex(globalData);
    retVal = sendControlCommand(globalData, cmd, 4, &ack, &actual);
    unlock_mutex(globalData);

    if (retVal < 0) {
        return retVal; 
    }

    if (ack != 0x20) {
        return ackStatus(ack); 
    }

    /* if the card is present, change the status to powered off */
    if (globalData->cards[socket].status)
	    globalData->cards[socket].status = 1; 

#ifdef ASE_DEBUG
    syslog(LOG_INFO, " CardPowerOff - Exit\n");
#endif

    return ASE_OK; 
}


/*****************************************************************************
*
*****************************************************************************/
/* on == 0x01, off == 0x00 */
int ChangeLedState (reader* globalData, char on) {
    char cmd[5], ack;
    int retVal, retVal2, actual;

#ifdef ASE_DEBUG
    syslog(LOG_INFO, "\n ChangeLedState - Enter\n");
#endif

    if (retVal = readerCommandInit(globalData, 1))
        return retVal;

    cmd[0] = ASE_PACKET_TYPE(0x50, globalData->commandCounter, 0);
    globalData->commandCounter++;
    globalData->commandCounter %= 4;
    cmd[1] = 0x17;
    cmd[2] = 0x01;
    cmd[3] = on;
    cmd[4] = cmd[0] ^ cmd[1] ^ cmd[2] ^ cmd[3];

    lock_mutex(globalData);
    retVal = sendControlCommand(globalData, cmd, 5, &ack, &actual);
    unlock_mutex(globalData);

    if (retVal < 0) {
        return retVal; 
    }

    if (ack != 0x20) {
        return ackStatus(ack); 
    }

#ifdef ASE_DEBUG
    syslog(LOG_INFO, " ChangeLedState - Exit\n");
#endif

    return ASE_OK; 
}


/*****************************************************************************
*
*****************************************************************************/
int EmptyOpenResponseProccessor   (reader* globalData, char socket, unsigned long arg, uchar* cs,
                                   char* outBuf, int* outBufLen) {
    return ASE_OK;
}

/*****************************************************************************
*
*****************************************************************************/
int T0CPUCardTransact (reader* globalData, char socket, char* inBuf, int len, char* outBuf, int* outBufLen,
                     openResponseProccessor orp, unsigned long arg) {
    char cmd[BULK_BUFFER_SIZE]; 
    uchar checksum;
    int retVal, retVal2, i, thisWrite;
    char* buffer = inBuf;

#ifdef ASE_DEBUG
    syslog(LOG_INFO, " T0CPUCardTransact - Enter\n");
#endif

    if (retVal = cardCommandInit(globalData, socket, 1))
        return retVal;

    while (len) {
        /* send at most T0_DATA_CHUNK_SIZE (==28) bytes of data in each packet */
        thisWrite = (len > T0_DATA_CHUNK_SIZE ? T0_DATA_CHUNK_SIZE : len);

        checksum = cmd[0] = ASE_PACKET_TYPE(0x50, globalData->commandCounter, socket);
        globalData->commandCounter++;
        globalData->commandCounter %= 4;
        cmd[1] = 0x28;
        cmd[2] = thisWrite;
        checksum ^= cmd[1];
        checksum ^= cmd[2];
        for (i = 0 ; i < thisWrite ; ++i) {
            cmd[i + 3] = buffer[i]; 
            checksum ^= cmd[i + 3];
        }
        cmd[i + 3] = checksum;

    	lock_mutex(globalData);
        if (len > T0_DATA_CHUNK_SIZE) {/* this is not the last packet */ 
            retVal = sendOpenResponseCommand(globalData, socket, cmd, thisWrite + 4, 
                                             outBuf, outBufLen, EmptyOpenResponseProccessor, arg, 0);
        }
        else { /* this is the last packet */
            retVal = sendOpenResponseCommand(globalData, socket, cmd, thisWrite + 4, 
                                             outBuf, outBufLen, orp, arg, 1);
        }
	    unlock_mutex(globalData);
    
        if (retVal < 0) {
            return retVal; 
        }

        len -= thisWrite;
	    buffer += thisWrite;
    }

#ifdef ASE_DEBUG
    syslog(LOG_INFO, " T0CPUCardTransact - Exit\n");
#endif

    return ASE_OK; 
}


/*****************************************************************************
*
*****************************************************************************/
int CPUCardTransact (reader* globalData, char socket, char* buffer, int len, char* outBuf, int* outBufLen,
                     openResponseProccessor orp, unsigned long arg) {
    char cmd[BULK_BUFFER_SIZE];
    uchar checksum;
    int retVal, retVal2, i;

#ifdef ASE_DEBUG
    syslog(LOG_INFO, " CPUCardTransact - Enter\n");
#endif

    if (retVal = cardCommandInit(globalData, socket, 1))
        return retVal;

    checksum = cmd[0] = ASE_PACKET_TYPE(0x50, globalData->commandCounter, socket);
    globalData->commandCounter++;
    globalData->commandCounter %= 4;
    cmd[1] = 0x28;
    cmd[2] = len;
    checksum ^= cmd[1];
    checksum ^= cmd[2];
    for (i = 0 ; i < len ; ++i) {
        cmd[i + 3] = buffer[i]; 
        checksum ^= cmd[i + 3];
    }
    cmd[len + 3] = checksum;

    lock_mutex(globalData);
    retVal = sendOpenResponseCommand(globalData, socket, cmd, len + 4, outBuf, outBufLen, orp, arg, 1);
    unlock_mutex(globalData);

    if (retVal < 0) {
        return retVal; 
    }

#ifdef ASE_DEBUG
    syslog(LOG_INFO, " CPUCardTransact - Exit\n");
#endif

    return ASE_OK; 
}



/*****************************************************************************
*
*****************************************************************************/
int T1CPUCardTransact (reader* globalData, char socket, 
                       char* buffer, int len, char* outBuf, int* outBufLen,
                       openResponseProccessor orp, unsigned long arg) {
    char cmd[9], ack;
    uchar checksum;
    char* data = &(buffer[3]);
    int retVal, retVal2, i, actual, numOfChunks, dataLen = buffer[2];

#ifdef ASE_DEBUG
    syslog(LOG_INFO, "\n T1CPUCardTransact - Enter\n");
#endif

    if (retVal = cardCommandInit(globalData, socket, 1))
        return retVal;

    /* send the header */
    cmd[0] = ASE_PACKET_TYPE(0x50, globalData->commandCounter, socket);
    globalData->commandCounter++;
    globalData->commandCounter %= 4;
    cmd[1] = 0x29;
    cmd[2] = len - dataLen; /* number of bytes in the header */
    cmd[3] = buffer[0]; /* NAD */
    cmd[4] = buffer[1]; /* PCB */
    cmd[5] = buffer[2]; /* LEN */

    checksum = cmd[0] ^ cmd[1] ^ cmd[2] ^ cmd[3] ^ cmd[4] ^ cmd[5];
    /* compute the checksum over all the information bytes as well */
    for (i = 0 ; i < buffer[2] ; ++i)
        checksum ^= buffer[3 + i];

    /* EDC */
    if (cmd[2] == 4) { /* LRC */
        cmd[6] = buffer[len - 1];
        checksum ^= cmd[6];
        cmd[7] = checksum;
    }
    else { /* CRC */
        cmd[6] = buffer[len - 2];
        cmd[7] = buffer[len - 1];
        checksum ^= cmd[6];
        checksum ^= cmd[7];
        cmd[8] = checksum;
    }

    /* send the command */
    lock_mutex(globalData);

    retVal = sendControlCommand(globalData, cmd, 4 + (len - buffer[2]), &ack, &actual);

    if (retVal < 0) {
        unlock_mutex(globalData);
        return retVal; 
    }

    if (ack != 0x20) {
        unlock_mutex(globalData);
        return ackStatus(ack); 
    }

    /* send the data (if any exists) */
    numOfChunks = (dataLen > T1_DATA_CHUNK_SIZE) ? 2 : 1;
    if (numOfChunks == 2) {
        /* the first chunk is sent and then ack byte should be received */

        retVal = writeToReader(globalData, data, T1_DATA_CHUNK_SIZE, &actual);
        
        if (retVal < 0 || actual != T1_DATA_CHUNK_SIZE) {
	        unlock_mutex(globalData);
            return retVal;
        }

        /* prepare for next chunk */
        data += T1_DATA_CHUNK_SIZE;
        dataLen -= T1_DATA_CHUNK_SIZE;

        /* read the ack byte */
        retVal = readCloseResponse(globalData, socket, 1, &ack, &actual);

        if (retVal < 0 || actual != 1) {
            unlock_mutex(globalData);
            return retVal;
        }

        if (ack != 0x20) {
            unlock_mutex(globalData);
            return ackStatus(ack); 
        }
    }

    /* send the last chunk of data as an open response command */
    retVal = sendOpenResponseCommand(globalData, socket, data, dataLen, outBuf, outBufLen, orp, arg, 1);

    unlock_mutex(globalData);

    if (retVal < 0) {
        return retVal; 
    }

#ifdef ASE_DEBUG
    syslog(LOG_INFO, " T1CPUCardTransact - Exit\n");
#endif

    return ASE_OK; 
}



/*****************************************************************************
*
*****************************************************************************/
int MemoryCardTransact (reader* globalData, char socket, 
                        char cmdType, uchar command, ushort address, uchar len,
                        uchar* data, char* outBuf, int* outBufLen) {
    char cmd[BULK_BUFFER_SIZE];
    uchar checksum, readByte;
    int retVal, i, actual;

#ifdef ASE_DEBUG
    syslog(LOG_INFO, "\n MemoryCardTransact - Enter\n");
#endif

    if (retVal = cardCommandInit(globalData, socket, 1))
        return retVal;

    cmd[0] = ASE_PACKET_TYPE(0x50, globalData->commandCounter, socket);
    globalData->commandCounter++;
    globalData->commandCounter %= 4;
    cmd[1] = 0x25;
    cmd[2] = 5 + (cmdType == 0 ? len : 0);
    cmd[3] = cmdType;
    cmd[4] = command;
    cmd[5] = (address >> 8) & 0xFF;
    cmd[6] = address & 0xFF;
    cmd[7] = len;
    checksum = cmd[0] ^ cmd[1] ^ cmd[2] ^ cmd[3] ^ cmd[4] ^ cmd[5] ^ cmd[6] ^ cmd[7];
    i = 0;
    if (cmdType == 0) {
        for (; i < len ; ++i) {
            cmd[i + 8] = data[i]; 
            checksum ^= cmd[i + 8];
        }
    }
    cmd[i + 8] = checksum;


    globalData->curPos = globalData->lastPos = 0;

    lock_mutex(globalData);

    /* send the command */
    retVal = writeToReader(globalData, cmd, cmd[2] + 4, &actual);
    if (retVal < 0) {
	    unlock_mutex(globalData);
        return retVal; 
    }

    checksum = 0;

    /* read the Ack control byte or the close response packet type */
    retVal = readCloseResponse(globalData, socket, 1, &readByte, &actual);
    if (retVal < 0) {
    	unlock_mutex(globalData);
        return retVal; 
    }

    checksum ^= readByte;

    /* if readByte == 0x2x */
    if ((readByte & 0xF0) == 0x20) {
        if (readByte != 0x20) {
/*            *outBufLen = 0; */
	        unlock_mutex(globalData);
            return ackStatus(readByte);
        }
    }
    /* ths response is a close response */
    else if ((readByte & 0xF0) == 0x10) {
        if (readByte != 0x10) {
	        unlock_mutex(globalData);
            return ackStatus(readByte); 
        }

        /* read data length */
        retVal = readCloseResponse(globalData, socket, 1, &readByte, &actual);
        if (retVal < 0) {
	        unlock_mutex(globalData);
            return retVal; 
        }

        checksum ^= readByte;

        /* read data  */
        retVal = readCloseResponse(globalData, socket, readByte, outBuf, outBufLen);
        if (retVal < 0 || (*outBufLen) != readByte) {
	        unlock_mutex(globalData);
            return retVal; 
        }

        for (i = 0 ; i < *outBufLen ; ++i)
            checksum ^= outBuf[i];

        /* read the checksum */
        retVal = readCloseResponse(globalData, socket, 1, &readByte, &actual);
	    
        if (retVal < 0) {
            unlock_mutex(globalData);
            return retVal; 
        }

        if (checksum != readByte) {
            unlock_mutex(globalData);
            return ASE_ERROR_CHECKSUM; 
        }
    }

    unlock_mutex(globalData);

#ifdef ASE_DEBUG
    syslog(LOG_INFO, " MemoryCardTransact - Exit\n\n");
#endif

    return ASE_OK; 
}


