#include "Ase.h"


/*****************************************************************************
*
*****************************************************************************/
#define MAX_T0_DATA_SIZE        300 



/*****************************************************************************
*
*****************************************************************************/
static int T0ReadOpenResponseProccessor   (reader* globalData, char socket, unsigned long arg, uchar* cs,
                                    char* outBuf, int* outBufLen) {
    int retVal, i, actual, p3, finish = 0;
    uchar procByte, ins, readByte;
    char stopLoop = 0;

    ins = (uchar)(arg & 0x000000FF);
    ins &= 0xFE;
    *outBufLen = p3 = (uchar)((arg & 0x0000FF00) >> 8);
    if (p3 == 0) {
        *outBufLen = p3 = 256;
    }

    /*
    syslog(LOG_INFO, "T0ReadOpenResponseProccessor: p3 = %d, ins = 0x%x%x, arg = 0x%x%x%x%x\n", p3, PRINT_CHAR(ins),
            PRINT_CHAR((uchar)((arg & 0x0000FF00) >> 8)), PRINT_CHAR((uchar)(arg & 0x000000FF)));
    */
    do {
    	/* get the next procedure byte */
        retVal = readOpenResponse(globalData, socket, 1, &readByte, &actual, 0);
        if (checkValidity(retVal, 1, actual, "T0ReadOpenResponseProccessor - Error! in procByte read.\n")) 
            return retVal; 
	
        (*cs) ^= readByte;
	
        procByte = (readByte & 0xFE);
	
    	/* Parse the procedure byte value */
	if (procByte == 0x60)   /* NULL byte */
	        continue;
    	/* Status byte SW1 */
	else if ((procByte & 0xf0) == 0x90 || (procByte & 0xf0) == 0x60) {
	    (*outBuf) = readByte;
	    outBuf++;
	    retVal = readOpenResponse (globalData, socket, 1, &readByte, &actual, 0);
	    if (checkValidity(retVal, 1, actual, "T0ReadOpenResponseProccessor - Error! in sw2 read.\n")) 
		return retVal; 
	    (*outBuf) = readByte;
	    outBuf++;
	    (*cs) ^= readByte;
	    stopLoop = 1;
	}
        /* read all remaining bytes */
	else if (procByte == ins) {
	    /*	    syslog(LOG_INFO, "T0ReadOpenResponseProccessor: procByte is ins\n"); */
            retVal = readOpenResponse (globalData, socket, p3, outBuf, &actual, 0);
            if (checkValidity(retVal, p3, actual, "T0ReadOpenResponseProccessor - Error! in all data read.\n")) 
                return retVal; 
            for (i = 0 ; i < actual ; i++) 
    	        (*cs) ^= outBuf[i];
	    p3 = 0;
	}
	/* read next byte */
	else if (procByte == (~ins & 0xFE)) {
	    /*	    syslog(LOG_INFO, "T0ReadOpenResponseProccessor: procByte is ~ins\n"); */
            retVal = readOpenResponse (globalData, socket, 1, outBuf++, &actual, 0);
            if (checkValidity(retVal, 1, actual, "T0ReadOpenResponseProccessor - Error! in one byte data read.\n"))
                return retVal; 
	    p3--;
	}
	else 
	    return ASE_ERROR_COMMAND; 
    } while (p3 && !stopLoop);
    
    /* set actual data read */
    *outBufLen = *outBufLen - p3; 
    
    if (!stopLoop) {
	/* Read SW1, SW2 */
	while (!finish) {
	    retVal = readOpenResponse (globalData, socket, 1, &readByte, &actual, 0);
	    if (checkValidity(retVal, 1, actual, "T0ReadOpenResponseProccessor - Error! in sw1 read.\n")) 
		return retVal; 
	    
	    if (readByte == 0x60)
		continue;

	    /*	syslog(LOG_INFO, "T0ReadOpenResponseProccessor: sw1 = 0x%x%x\n", PRINT_CHAR(readByte)); */
	    (*cs) ^= readByte;
	    outBuf[*outBufLen] = readByte;
	    (*outBufLen)++;
	    finish = 1;
	}
	
        retVal = readOpenResponse (globalData, socket, 1, &readByte, &actual, 0);
        if (checkValidity(retVal, 1, actual, "T0ReadOpenResponseProccessor - Error! in sw22 read.\n")) 
            return retVal; 
	/*	syslog(LOG_INFO, "T0ReadOpenResponseProccessor: sw1 = 0x%x%x\n", PRINT_CHAR(readByte)); */
        (*cs) ^= readByte;
        outBuf[*outBufLen] = readByte;
        (*outBufLen)++;
    }

    return ASE_OK;
}




/*****************************************************************************
*
*****************************************************************************/
static int T0Read1  (reader* globalData, uchar socket, 
             uchar cla, uchar ins, uchar p1, uchar p2, int* dataLength,
             uchar* dataBuffer, uchar* sw1, uchar* sw2) {
    int retVal, i, p3;
    unsigned long arg;
    char outBuffer[MAX_T0_DATA_SIZE], Header[5];


#ifdef ASE_DEBUG
    syslog(LOG_INFO, "\n T0Read - Enter\n============\n");
#endif

    if (retVal = cardCommandInit(globalData, socket, 1))
        return retVal;

    p3 = *dataLength;

    arg = ins;
    arg &= 0x000000FF;
    arg |= ((p3 << 8) & 0x0000FF00);

    /*
    syslog(LOG_INFO, "T0Read: p3 = %d, ins = 0x%x%x, arg = 0x%x%x%x%x\n", p3, PRINT_CHAR(ins),
            PRINT_CHAR((uchar)((arg & 0x0000FF00) >> 8)), PRINT_CHAR((uchar)(arg & 0x000000FF)));
    */

    Header[0] = cla;
    Header[1] = ins;
    Header[2] = p1;
    Header[3] = p2;
    Header[4] = p3;

    retVal = T0CPUCardTransact(globalData, socket, Header, 5, outBuffer, dataLength, 
                             T0ReadOpenResponseProccessor, arg);
    if (retVal < 0) {
        return retVal; 
    }

    i = 0;
    if (*dataLength >= 2) {
        /* copy the data */
        for ( ; i < (*dataLength - 2) ; ++i)
            dataBuffer[i] = outBuffer[i];
    }
    *sw1 = outBuffer[i++];
    *sw2 = outBuffer[i++];

    if (*dataLength >= 2)
        (*dataLength) -= 2; // dont include the sw1+sw2

#ifdef ASE_DEBUG
    syslog(LOG_INFO, "T0Read - sw1 = 0x%x sw2 = 0x%x\n", *sw1, *sw2);
    syslog(LOG_INFO, " T0Read - Exit\n");
#endif

    return ASE_OK; 
}



/*****************************************************************************
*
*****************************************************************************/
static int T0WriteOpenResponseProccessor  (reader* globalData, char socket, unsigned long arg, uchar* cs,
                                    char* outBuf, int* outBufLen) {
    int retVal, i, actual, finish = 0;
    uchar readByte1, readByte2;

    while (!finish) {
	retVal = readOpenResponse(globalData, socket, 1, &readByte1, &actual, 0);
	if (retVal < 0 || actual != 1) {
	    return retVal; 
	}
	*outBufLen = 1;
	
	/* if the read byte is sw1, then read sw2 as well */
	if (readByte1 == 0x60) /* NULL byte */
	    continue; /* do nothing */
	else if ((readByte1 & 0xF0) == 0x90 || (readByte1 & 0xF0) == 0x60) {
	    retVal = readOpenResponse(globalData, socket, 1, &readByte2, &actual, 0);
	    /*	syslog(LOG_INFO, "T0WriteOpenResponseProccessor: readByte2 = 0x%x%x\n", PRINT_CHAR(readByte2)); */
	    if (checkValidity(retVal, 1, actual, "T0WriteOpenResponseProccessor - Error! in second readByte read.\n")) 
		return retVal; 
	    *outBufLen = 2;
	}
	
	finish = 1;
    }

    outBuf[0] = readByte1;
    (*cs) ^= outBuf[0];
    if (*outBufLen == 2) {
	outBuf[1] = readByte2;
	(*cs) ^= outBuf[1];
    }
    
    return ASE_OK;
}



/*****************************************************************************
*
*****************************************************************************/
static int T0Write1  (reader* globalData, uchar socket, 
              uchar cla, uchar ins, uchar p1, uchar p2, uchar p3,
              uchar* dataBuffer, uchar* sw1, uchar* sw2) {
    int retVal, i, actual;
    uchar procByte, readBytes[2];
    char Header[5];
    int stopLoop = 0;


#ifdef ASE_DEBUG
    syslog(LOG_INFO, "\n T0Write - Enter\n================\n");
#endif

    if (retVal = cardCommandInit(globalData, socket, 1))
        return retVal;

    ins &= 0xFE;

    /*    syslog(LOG_INFO, "T0Write - ins = 0x%x, ~ins = 0x%x\n", ins, (~ins & 0xFE)); */

    Header[0] = cla;
    Header[1] = ins;
    Header[2] = p1;
    Header[3] = p2;
    Header[4] = p3;

    /* get the next procedure byte */
    retVal = T0CPUCardTransact(globalData, socket, Header, 5, readBytes, &actual, 
			     T0WriteOpenResponseProccessor, 0);
    if (retVal < 0) {
	return retVal; 
    }
    
    do {
        procByte = (readBytes[0] & 0xFE);
	/*		syslog(LOG_INFO, "T0Write: procByte = 0x%x%x\n", PRINT_CHAR(procByte)); */

        /* Parse the procedure byte value */
        if (procByte == 0x60) {  /* NULL byte */
            continue;
	}
        /* Status byte SW1 */
        else if ((procByte & 0xF0) == 0x90 || (procByte & 0xF0) == 0x60) {
	    *sw1 = readBytes[0];
            *sw2 = readBytes[1];
	    /*	    syslog(LOG_INFO, "T0Write: sw1 = 0x%x%x sw2 = 0x%x%x\n", PRINT_CHAR(*sw1), PRINT_CHAR(*sw2)); */
            stopLoop = 1;
        }

        /* write all remaining bytes */
	else if (procByte == ins) {
	    /*	    	    syslog(LOG_INFO, "T0Write: procByte is equal to ins\n"); */
            retVal = T0CPUCardTransact(globalData, socket, dataBuffer, p3, readBytes, &actual, 
                                     T0WriteOpenResponseProccessor, 0);
            if (retVal < 0) {
                return retVal; 
            }
	    p3 = 0;
	}
	/* write next byte */
	else if (procByte == (~ins & 0xFE)) {
	    /*	    syslog(LOG_INFO, "T0Write: procByte is equal to ~ins\n"); */
            retVal = T0CPUCardTransact(globalData, socket, dataBuffer++, 1, readBytes, &actual, 
                                     T0WriteOpenResponseProccessor, 0);
            if (retVal < 0) {
                return retVal; 
            }
	    p3--;
	}
        else 
            return ASE_ERROR_COMMAND; 
    } while (p3 && !stopLoop);
    
    if (!stopLoop) {
        /* Read SW1, SW2 */
        *sw1 = readBytes[0];
        *sw2 = readBytes[1];
    }

#ifdef ASE_DEBUG
    syslog(LOG_INFO, " T0Write - Exit\n");
#endif

    return ASE_OK;
}





/*****************************************************************************
*
*****************************************************************************/
/* outBuf will contain the SW1/SW2 as well */
int T0Read  (reader* globalData, uchar socket, 
             uchar* inBuf, int inBufLen, uchar* outBuf, int *outBufLen) {
    uchar cla, ins, p1, p2, sw1, sw2;
    int retVal;

    if (inBufLen < 5)
        return ASE_ERROR_COMMAND;

    cla = inBuf[0];
    ins = inBuf[1];
    p1 = inBuf[2];
    p2 = inBuf[3];
    *outBufLen = inBuf[4];

    retVal = T0Read1(globalData, socket, cla, ins, p1, p2, outBufLen, outBuf, &sw1, &sw2);
    if (retVal < 0)
        return retVal;

    /* Assuming outBuf is large enough to hold data + SW1, SW2 */
    outBuf[*outBufLen] = sw1;
    (*outBufLen)++;
    outBuf[*outBufLen] = sw2;
    (*outBufLen)++;

#ifdef ASE_DEBUG
    syslog(LOG_INFO, "T0Read: SW1 = 0x%x%x SW2 = 0x%x%x actual = %d\n", PRINT_CHAR(sw1), PRINT_CHAR(sw2), *outBufLen);
#endif

    return ASE_OK;
}


/*****************************************************************************
*
*****************************************************************************/
int T0Write  (reader* globalData, uchar socket, 
              uchar* inBuf, int inBufLen, uchar* outBuf, int *outBufLen) {
    uchar cla, ins, p1, p2, p3, sw1, sw2;
    int retVal;

    if (inBufLen < 5)
        return ASE_ERROR_COMMAND;

    cla = inBuf[0];
    ins = inBuf[1];
    p1 = inBuf[2];
    p2 = inBuf[3];
    p3 = inBuf[4];

    retVal = T0Write1(globalData, socket, cla, ins, p1, p2, p3, &(inBuf[5]), &sw1, &sw2);
    if (retVal < 0)
        return retVal;

    /* Assuming outBuf is large enough to hold SW1, SW2 */
    *outBufLen = 2;
    outBuf[0] = sw1;
    outBuf[1] = sw2;

#ifdef ASE_DEBUG
    syslog(LOG_INFO, "T0Write: SW1 = 0x%x%x SW2 = 0x%x%x\n", PRINT_CHAR(sw1), PRINT_CHAR(sw2));
#endif

    return ASE_OK;
}


